@@ -1,6 +1,8 @@
+use std::collections::HashSet;
 use std::error::Error;
-use std::fs;
+use std::fs::{self, File};
 use std::io;
+use std::io::prelude::*;
 use std::path::{Path, PathBuf};
 use std::rc::Rc;
 
@@ -21,30 +23,108 @@ use crate::consts;
 mod error;
 use error::Error as FDroidError;
 
-pub async fn download_apps(
-    app_ids: Vec<String>,
-    parallel: usize,
-    sleep_duration: u64,
-    outpath: &PathBuf,
-) {
-    let dir = match tempdir() {
-        Ok(dir) => dir,
+async fn retrieve_index_or_exit() -> Value {
+    let temp_dir = match tempdir() {
+        Ok(temp_dir) => temp_dir,
         Err(_) => {
             println!("Could not create temporary directory for F-Droid package index. Exiting.");
             std::process::exit(1);
         }
     };
-    let files = download_and_extract_index_to_tempdir(&dir).await;
-    let index = match verify_and_return_index(&dir, &files) {
-        Ok(index) => index,
-        Err(_) => {
-            println!("Could not verify F-Droid package index. Exiting.");
+    let config_dir = match dirs::config_dir() {
+        Some(mut config_dir) => {
+            let create_dir = |config_dir: &PathBuf| {
+                if !config_dir.is_dir() &&  fs::create_dir(config_dir).is_err() {
+                    println!("Could not create a config directory for apkeep to store F-Droid package index. Exiting.");
+                    std::process::exit(1);
+                }
+            };
+            create_dir(&config_dir);
+            config_dir.push("apkeep");
+            create_dir(&config_dir);
+            config_dir
+        },
+        None => {
+            println!("Could not find a config directory for apkeep to store F-Droid package index. Exiting.");
             std::process::exit(1);
         },
     };
-    
-    let (apps, repo_address) = match parse_json(index, app_ids) {
-        Ok((apps, repo_address)) => (apps, repo_address),
+    let mut latest_etag_file = PathBuf::from(&config_dir);
+    latest_etag_file.push("latest_etag");
+    let latest_etag = match File::open(&latest_etag_file) {
+        Ok(mut file) => {
+            let mut contents = String::new();
+            if file.read_to_string(&mut contents).is_err() {
+                println!("Could not read latest_etag file for F-Droid package index. Exiting.");
+                std::process::exit(1);
+            }
+            Some(contents)
+        },
+        Err(_) => None,
+    };
+
+    let http_client = reqwest::Client::new();
+    let index_response = http_client
+        .head(consts::FDROID_INDEX_URL)
+        .send().await.unwrap();
+
+    let etag = if index_response.headers().contains_key("ETag") {
+        index_response.headers()["ETag"].to_str().unwrap()
+    } else {
+        println!("Could not receive etag for F-Droid package index. Exiting.");
+        std::process::exit(1);
+    };
+
+    let mut index_file = PathBuf::from(&config_dir);
+    index_file.push("index_v1.json");
+    if latest_etag.is_some() && latest_etag.unwrap() == etag {
+        let index = {
+            let mut file = File::open(&index_file).unwrap();
+            let mut contents = String::new();
+            file.read_to_string(&mut contents).unwrap();
+            contents
+        };
+        serde_json::from_str(&index).unwrap()
+    } else {
+        let files = download_and_extract_index_to_tempdir(&temp_dir).await;
+        match verify_and_return_index(&temp_dir, &files) {
+            Ok(index) => {
+                match serde_json::from_str(&index) {
+                    Ok(index_value) => {
+                        if fs::write(index_file, index).is_err() {
+                            println!("Could not write F-Droid package index to config file. Exiting.");
+                            std::process::exit(1);
+                        }
+                        if fs::write(latest_etag_file, etag).is_err() {
+                            println!("Could not write F-Droid etag to config file. Exiting.");
+                            std::process::exit(1);
+                        }
+                        index_value
+                    }
+                    Err(_) => {
+                        println!("Could not decode JSON for F-Droid package index. Exiting.");
+                        std::process::exit(1);
+                    }
+                }
+            },
+            Err(_) => {
+                println!("Could not verify F-Droid package index. Exiting.");
+                std::process::exit(1);
+            },
+        }
+    }
+}
+
+pub async fn download_apps(
+    apps: Vec<(String, Option<String>)>,
+    parallel: usize,
+    sleep_duration: u64,
+    outpath: &Path,
+) {
+    let index = retrieve_index_or_exit().await;
+
+    let (fdroid_apps, repo_address) = match parse_json_for_download_information(index, apps) {
+        Ok((fdroid_apps, repo_address)) => (fdroid_apps, repo_address),
         Err(_) => {
             println!("Could not parse JSON of F-Droid package index. Exiting.");
             std::process::exit(1);
@@ -53,32 +133,45 @@ pub async fn download_apps(
 
     let repo_address = Rc::new(repo_address);
     futures_util::stream::iter(
-        apps.into_iter().map(|app| {
-            let (app_id, url_filename, hash) = app;
+        fdroid_apps.into_iter().map(|fdroid_app| {
+            let (app_id, app_version, url_filename, hash) = fdroid_app;
             let repo_address = Rc::clone(&repo_address);
             async move {
-                println!("Downloading {}...", app_id);
+                let app_string = match app_version {
+                    Some(ref version) => {
+                        println!("Downloading {} version {}...", app_id, version);
+                        format!("{}@{}", app_id, version)
+                    },
+                    None => {
+                        println!("Downloading {}...", app_id);
+                        app_id.to_string()
+                    },
+                };
+                let fname = format!("{}.apk", app_string);
                 if sleep_duration > 0 {
                     sleep(Duration::from_millis(sleep_duration)).await;
                 }
                 let download_url = format!("{}/{}", repo_address, url_filename);
-                let fname = format!("{}.apk", app_id);
-                let sha256sum = match tokio_dl_stream_to_disk::download_and_return_sha256sum(&download_url, &Path::new(outpath), &fname).await {
+                let sha256sum = match tokio_dl_stream_to_disk::download_and_return_sha256sum(&download_url, Path::new(outpath), &fname).await {
                     Ok(sha256sum) => Some(sha256sum),
                     Err(err) if matches!(err.kind(), TDSTDErrorKind::FileExists) => {
-                        println!("File already exists for {}. Skipping...", app_id);
+                        println!("File already exists for {}. Skipping...", app_string);
+                        None
+                    },
+                    Err(err) if matches!(err.kind(), TDSTDErrorKind::PermissionDenied) => {
+                        println!("Permission denied when attempting to write file for {}. Skipping...", app_string);
                         None
                     },
                     Err(_) => {
-                        println!("An error has occurred attempting to download {}.  Retry #1...", app_id);
-                        match tokio_dl_stream_to_disk::download_and_return_sha256sum(&download_url, &Path::new(outpath), &fname).await {
+                        println!("An error has occurred attempting to download {}.  Retry #1...", app_string);
+                        match tokio_dl_stream_to_disk::download_and_return_sha256sum(&download_url, Path::new(outpath), &fname).await {
                             Ok(sha256sum) => Some(sha256sum),
                             Err(_) => {
-                                println!("An error has occurred attempting to download {}.  Retry #2...", app_id);
-                                match tokio_dl_stream_to_disk::download_and_return_sha256sum(&download_url, &Path::new(outpath), &fname).await {
+                                println!("An error has occurred attempting to download {}.  Retry #2...", app_string);
+                                match tokio_dl_stream_to_disk::download_and_return_sha256sum(&download_url, Path::new(outpath), &fname).await {
                                     Ok(sha256sum) => Some(sha256sum),
                                     Err(_) => {
-                                        println!("An error has occurred attempting to download {}. Skipping...", app_id);
+                                        println!("An error has occurred attempting to download {}. Skipping...", app_string);
                                         None
                                     }
                                 }
@@ -88,9 +181,9 @@ pub async fn download_apps(
                 };
                 if let Some(sha256sum) = sha256sum {
                     if sha256sum == hash {
-                        println!("{} downloaded successfully!", app_id);
+                        println!("{} downloaded successfully!", app_string);
                     } else {
-                        println!("{} downloaded, but the sha256sum does not match the one signed by F-Droid. Proceed with caution.", app_id);
+                        println!("{} downloaded, but the sha256sum does not match the one signed by F-Droid. Proceed with caution.", app_string);
                     }
                 }
             }
@@ -98,7 +191,8 @@ pub async fn download_apps(
     ).buffer_unordered(parallel).collect::<Vec<()>>().await;
 }
 
-fn parse_json(index: Value, app_ids: Vec<String>) -> Result<(Vec<(String, String, Vec<u8>)>, String), FDroidError> { //(Vec<(String, String)>, String), FDroidError> {
+type DownloadInformation = (Vec<(String, Option<String>, String, Vec<u8>)>, String);
+fn parse_json_for_download_information(index: Value, apps: Vec<(String, Option<String>)>) -> Result<DownloadInformation, FDroidError> {
     let index_map = index.as_object().ok_or(FDroidError::Dummy)?;
     let repo_address = index_map
         .get("repo").ok_or(FDroidError::Dummy)?
@@ -109,45 +203,117 @@ fn parse_json(index: Value, app_ids: Vec<String>) -> Result<(Vec<(String, String
         .get("packages").ok_or(FDroidError::Dummy)?
         .as_object().ok_or(FDroidError::Dummy)?;
 
-    let apps: Vec<(String, String, Vec<u8>)> = app_ids.into_iter().map(|app_id| {
+    let fdroid_apps: Vec<(String, Option<String>, String, Vec<u8>)> = apps.into_iter().map(|app| {
+        let (app_id, app_version) = app;
         if packages.contains_key(&app_id) {
             let app_array_value = packages.get(&app_id).unwrap();
             if app_array_value.is_array() {
                 let app_array = app_array_value.as_array().unwrap();
-                if app_array.len() >= 1 && app_array[0].is_object() {
-                    let app = app_array[0].as_object().unwrap();
-                    if app.contains_key("apkName") && app.contains_key("hash") {
-                        let filename_value = app.get("apkName").unwrap();
-                        let hash_value = app.get("hash").unwrap();
-                        if filename_value.is_string() && hash_value.is_string() {
-                            let filename = filename_value.as_str().unwrap().to_string();
-                            if let Ok(hash) = hex::decode(hash_value.as_str().unwrap().to_string()) {
-                                return Some((app_id.to_string(), filename, hash));
+                if app_version.is_none() {
+                    if !app_array.is_empty() && app_array[0].is_object() {
+                        let fdroid_app = app_array[0].as_object().unwrap();
+                        if fdroid_app.contains_key("apkName") && fdroid_app.contains_key("hash") {
+                            let filename_value = fdroid_app.get("apkName").unwrap();
+                            let hash_value = fdroid_app.get("hash").unwrap();
+                            if filename_value.is_string() && hash_value.is_string() {
+                                let filename = filename_value.as_str().unwrap().to_string();
+                                if let Ok(hash) = hex::decode(hash_value.as_str().unwrap().to_string()) {
+                                    return Some((app_id, app_version, filename, hash));
+                                }
                             }
                         }
                     }
+                } else {
+                    for single_app in app_array {
+                        if single_app.is_object() {
+                            let fdroid_app = single_app.as_object().unwrap();
+                            if fdroid_app.contains_key("versionName") {
+                                let version_name_value = fdroid_app.get("versionName").unwrap();
+                                if version_name_value.is_string() {
+                                    let version_name = version_name_value.as_str().unwrap().to_string();
+                                    if version_name == *app_version.as_ref().unwrap() && fdroid_app.contains_key("apkName") && fdroid_app.contains_key("hash") {
+                                        let filename_value = fdroid_app.get("apkName").unwrap();
+                                        let hash_value = fdroid_app.get("hash").unwrap();
+                                        if filename_value.is_string() && hash_value.is_string() {
+                                            let filename = filename_value.as_str().unwrap().to_string();
+                                            if let Ok(hash) = hex::decode(hash_value.as_str().unwrap().to_string()) {
+                                                return Some((app_id, app_version, filename, hash));
+                                            }
+                                        }
+                                    }
+                                }
+                            }
+                        }
+                    }
+                    println!("Could not find version {} of {}. Skipping...", app_version.unwrap(), app_id);
+                    return None;
                 }
             }
         } else {
             println!("Could not find {} in package list. Skipping...", app_id);
         }
         None
-    }).filter_map(|i| i).collect();
+    }).flatten().collect();
 
-    Ok((apps, repo_address.to_string()))
+    Ok((fdroid_apps, repo_address.to_string()))
 }
 
-fn verify_and_return_index(dir: &TempDir, files: &Vec<String>) -> Result<Value, Box<dyn Error>> {
+pub async fn list_versions(apps: Vec<(String, Option<String>)>) {
+    let index = retrieve_index_or_exit().await;
+    if parse_json_display_versions(index, apps).is_err() {
+        println!("Could not parse JSON of F-Droid package index. Exiting.");
+    };
+}
+
+fn parse_json_display_versions(index: Value, apps: Vec<(String, Option<String>)>) -> Result<(), FDroidError> {
+    let index_map = index.as_object().ok_or(FDroidError::Dummy)?;
+
+    let packages = index_map
+        .get("packages").ok_or(FDroidError::Dummy)?
+        .as_object().ok_or(FDroidError::Dummy)?;
+
+    for app in apps {
+        let (app_id, _) = app;
+        println!("Versions available for {} on F-Droid:", app_id);
+        if packages.contains_key(&app_id) {
+            let app_array_value = packages.get(&app_id).unwrap();
+            if app_array_value.is_array() {
+                let app_array = app_array_value.as_array().unwrap();
+                let mut versions = HashSet::new();
+                for single_app in app_array {
+                    if single_app.is_object() {
+                        let fdroid_app = single_app.as_object().unwrap();
+                        if fdroid_app.contains_key("versionName") {
+                            let version_name_value = fdroid_app.get("versionName").unwrap();
+                            if version_name_value.is_string() {
+                                let version_name = version_name_value.as_str().unwrap().to_string();
+                                versions.insert(version_name);
+                            }
+                        }
+                    }
+                }
+                let mut versions = versions.drain().collect::<Vec<String>>();
+                versions.sort();
+                println!("| {}", versions.join(", "));
+            }
+        } else {
+            println!("| Could not find {} in package list. Skipping...", app_id);
+        }
+    }
+    Ok(())
+}
+
+fn verify_and_return_index(dir: &TempDir, files: &[String]) -> Result<String, Box<dyn Error>> {
     println!("Verifying...");
     let re = Regex::new(consts::FDROID_SIGNATURE_BLOCK_FILE_REGEX).unwrap();
     let cert_file = {
         let mut cert_files = vec![];
         for file in files {
-            if re.is_match(&file) {
+            if re.is_match(file) {
                 cert_files.push(file.clone());
             }
         }
-        if cert_files.len() < 1 {
+        if cert_files.is_empty() {
             return Err(Box::new(SimpleError::new("Found no certificate file for F-Droid repository.")));
         }
         if cert_files.len() > 1 {
@@ -178,7 +344,7 @@ fn verify_and_return_index(dir: &TempDir, files: &Vec<String>) -> Result<Value,
     let manifest_file_data = fs::read(manifest_file)?;
     {
         let signed_file_regex = Regex::new(r"\r\nSHA1-Digest-Manifest: (.*)\r\n").unwrap();
-        let signed_file_manifest_sha1sum = base64::decode(match signed_file_regex.captures(&signed_file_string) {
+        let signed_file_manifest_sha1sum = base64::decode(match signed_file_regex.captures(signed_file_string) {
             Some(caps) if caps.len() >= 2 => caps.get(1).unwrap().as_str(),
             _ => {
                 return Err(Box::new(SimpleError::new("Could not retrieve the manifest sha1sum from the signed file.")));
@@ -197,7 +363,7 @@ fn verify_and_return_index(dir: &TempDir, files: &Vec<String>) -> Result<Value,
     let index_file_data = fs::read(index_file)?;
     {
         let manifest_file_regex = Regex::new(r"\r\nName: index-v1\.json\r\nSHA1-Digest: (.*)\r\n").unwrap();
-        let manifest_file_index_sha1sum = base64::decode(match manifest_file_regex.captures(&manifest_file_string) {
+        let manifest_file_index_sha1sum = base64::decode(match manifest_file_regex.captures(manifest_file_string) {
             Some(caps) if caps.len() >= 2 => caps.get(1).unwrap().as_str(),
             _ => {
                 return Err(Box::new(SimpleError::new("Could not retrieve the index sha1sum from the manifest file.")));
@@ -211,7 +377,7 @@ fn verify_and_return_index(dir: &TempDir, files: &Vec<String>) -> Result<Value,
         }
     }
 
-    Ok(serde_json::from_str(std::str::from_utf8(&index_file_data)?)?)
+    Ok(String::from(std::str::from_utf8(&index_file_data)?))
 }
 
 fn get_certificate_and_signer_info_from_signature_block_file(signature_block_file: PathBuf) -> Result<(CapturedX509Certificate, SignerInfo), Box<dyn Error>> {
@@ -222,14 +388,14 @@ fn get_certificate_and_signer_info_from_signature_block_file(signature_block_fil
             if certificates.len() > 1 {
                 return Err(Box::new(SimpleError::new("Too many certificates provided.")));
             }
-            if certificates.len() < 1 {
+            if certificates.is_empty() {
                 return Err(Box::new(SimpleError::new("No certificate provided.")));
             }
             let signatories: Vec<&SignerInfo> = signed_data.signers().collect();
             if signatories.len() > 1 {
                 return Err(Box::new(SimpleError::new("Too many signatories provided.")));
             }
-            if signatories.len() < 1 {
+            if signatories.is_empty() {
                 return Err(Box::new(SimpleError::new("No signatories provided.")));
             }
             Ok((certificates[0].clone(), signatories[0].clone()))
